@@ -1,5 +1,5 @@
 # Operations.py
-# Author: Marcus D. Bloice <https://github.com/mdbloice>
+# Author: Marcus D. Bloice <https://github.com/mdbloice> and contributors
 # Licensed under the terms of the MIT Licence.
 """
 The Operations module contains classes for all operations used by Augmentor.
@@ -24,7 +24,7 @@ from __future__ import (absolute_import, division,
                         print_function, unicode_literals)
 from builtins import *
 
-from PIL import Image, ImageOps
+from PIL import Image, ImageOps, ImageEnhance
 import math
 from math import floor, ceil
 
@@ -266,6 +266,159 @@ class BlackAndWhite(Operation):
         return augmented_images
 
 
+class RandomBrightness(Operation):
+    """
+    This class is used to random change image brightness.
+    """
+    def __init__(self, probability, min_factor, max_factor):
+        """
+        required :attr:`probability` parameter
+
+        :func:`~Augmentor.Pipeline.Pipeline.random_brightness` function.
+
+        :param probability: Controls the probability that the operation is
+         performed when it is invoked in the pipeline.
+        :param min_factor: The value between 0.0 and max_factor
+         that define the minimum adjustment of image brightness.
+         The value  0.0 gives a black image,The value 1.0 gives the original image, value bigger than 1.0 gives more bright image.
+        :param max_factor: A value should be bigger than min_factor.
+         that define the maximum adjustment of image brightness.
+         The value  0.0 gives a black image, value 1.0 gives the original image, value bigger than 1.0 gives more bright image.
+
+        :type probability: Float
+        :type max_factor: Float
+        :type max_factor: Float
+        """
+        Operation.__init__(self, probability)
+        self.min_factor = min_factor
+        self.max_factor = max_factor
+
+    def perform_operation(self, images):
+        """
+        Random change the passed image brightness.
+
+        :param images: The image to convert into monochrome.
+        :type images: List containing PIL.Image object(s).
+        :return: The transformed image(s) as a list of object(s) of type
+         PIL.Image.
+        """
+        factor = np.random.uniform(self.min_factor, self.max_factor)
+
+        def do(image):
+
+            image_enhancer_brightness = ImageEnhance.Brightness(image)
+            return image_enhancer_brightness.enhance(factor)
+
+        augmented_images = []
+
+        for image in images:
+            augmented_images.append(do(image))
+
+        return augmented_images
+
+
+class RandomColor(Operation):
+    """
+    This class is used to random change saturation of an image.
+    """
+    def __init__(self, probability, min_factor, max_factor):
+        """
+        required :attr:`probability` parameter
+
+        :func:`~Augmentor.Pipeline.Pipeline.random_color` function.
+
+        :param probability: Controls the probability that the operation is
+         performed when it is invoked in the pipeline.
+        :param min_factor: The value between 0.0 and max_factor
+         that define the minimum adjustment of image saturation.
+         The value 0.0 gives a black and white image, value 1.0 gives the original image.
+        :param max_factor: A value should be bigger than min_factor.
+         that define the maximum adjustment of image saturation.
+         The value 0.0 gives a black and white image, value 1.0 gives the original image.
+
+        :type probability: Float
+        :type max_factor: Float
+        :type max_factor: Float
+        """
+        Operation.__init__(self, probability)
+        self.min_factor = min_factor
+        self.max_factor = max_factor
+
+    def perform_operation(self, images):
+        """
+        Random change the passed image saturation.
+
+        :param images: The image to convert into monochrome.
+        :type images: List containing PIL.Image object(s).
+        :return: The transformed image(s) as a list of object(s) of type
+         PIL.Image.
+        """
+        factor = np.random.uniform(self.min_factor, self.max_factor)
+
+        def do(image):
+
+            image_enhancer_color = ImageEnhance.Color(image)
+            return image_enhancer_color.enhance(factor)
+
+        augmented_images = []
+
+        for image in images:
+            augmented_images.append(do(image))
+
+        return augmented_images
+
+
+class RandomContrast(Operation):
+    """
+    This class is used to random change contrast of an image.
+    """
+    def __init__(self, probability, min_factor,max_factor):
+        """
+        required :attr:`probability` parameter
+
+        :func:`~Augmentor.Pipeline.Pipeline.random_contrast` function.
+
+        :param probability: Controls the probability that the operation is
+         performed when it is invoked in the pipeline.
+        :param min_factor: The value between 0.0 and max_factor
+         that define the minimum adjustment of image contrast.
+         The value  0.0 gives s solid grey image, value 1.0 gives the original image.
+        :param max_factor: A value should be bigger than min_factor.
+         that define the maximum adjustment of image contrast.
+         The value  0.0 gives s solid grey image, value 1.0 gives the original image.
+
+        :type probability: Float
+        :type max_factor: Float
+        :type max_factor: Float
+        """
+        Operation.__init__(self, probability)
+        self.min_factor = min_factor
+        self.max_factor = max_factor
+
+    def perform_operation(self, images):
+        """
+        Random change the passed image contrast.
+
+        :param images: The image to convert into monochrome.
+        :type images: List containing PIL.Image object(s).
+        :return: The transformed image(s) as a list of object(s) of type
+         PIL.Image.
+        """
+        factor = np.random.uniform(self.min_factor, self.max_factor)
+
+        def do(image):
+
+            image_enhancer_contrast = ImageEnhance.Contrast(image)
+            return image_enhancer_contrast.enhance(factor)
+
+        augmented_images = []
+
+        for image in images:
+            augmented_images.append(do(image))
+
+        return augmented_images
+
+
 class Skew(Operation):
     """
     This class is used to perform perspective skewing on images. It allows
@@ -481,7 +634,7 @@ class RotateStandard(Operation):
     .. seealso:: For 90 degree rotations, see the :class:`Rotate` class.
     """
 
-    def __init__(self, probability, max_left_rotation, max_right_rotation, expand=False):
+    def __init__(self, probability, max_left_rotation, max_right_rotation, expand=False, fillcolor=None):
         """
         Documentation to appear.
         """
@@ -489,6 +642,7 @@ class RotateStandard(Operation):
         self.max_left_rotation = -abs(max_left_rotation)   # Ensure always negative
         self.max_right_rotation = abs(max_right_rotation)  # Ensure always positive
         self.expand = expand
+        self.fillcolor = fillcolor
 
     def perform_operation(self, images):
         """
@@ -512,7 +666,7 @@ class RotateStandard(Operation):
             rotation = random_right
 
         def do(image):
-            return image.rotate(rotation, expand=self.expand, resample=Image.BICUBIC)
+            return image.rotate(rotation, expand=self.expand, resample=Image.BICUBIC, fillcolor=self.fillcolor)
 
         augmented_images = []
 
@@ -847,9 +1001,12 @@ class Crop(Operation):
          PIL.Image.
         """
 
-        def do(image):
+        w, h = images[0].size  # All images must be the same size, so we can just check the first image in the list
 
-            w, h = image.size
+        left_shift = random.randint(0, int((w - self.width)))
+        down_shift = random.randint(0, int((h - self.height)))
+
+        def do(image):
 
             # TODO: Fix. We may want a full crop.
             if self.width > w or self.height > h:
@@ -858,8 +1015,6 @@ class Crop(Operation):
             if self.centre:
                 return image.crop(((w/2)-(self.width/2), (h/2)-(self.height/2), (w/2)+(self.width/2), (h/2)+(self.height/2)))
             else:
-                left_shift = random.randint(0, int((w - self.width)))
-                down_shift = random.randint(0, int((h - self.height)))
                 return image.crop((left_shift, down_shift, self.width + left_shift, self.height + down_shift))
 
         augmented_images = []
@@ -1297,40 +1452,40 @@ class Distort(Operation):
             if i not in last_row and i not in last_column:
                 polygon_indices.append([i, i + 1, i + horizontal_tiles, i + 1 + horizontal_tiles])
 
+        for a, b, c, d in polygon_indices:
+            dx = random.randint(-self.magnitude, self.magnitude)
+            dy = random.randint(-self.magnitude, self.magnitude)
+
+            x1, y1, x2, y2, x3, y3, x4, y4 = polygons[a]
+            polygons[a] = [x1, y1,
+                            x2, y2,
+                            x3 + dx, y3 + dy,
+                            x4, y4]
+
+            x1, y1, x2, y2, x3, y3, x4, y4 = polygons[b]
+            polygons[b] = [x1, y1,
+                            x2 + dx, y2 + dy,
+                            x3, y3,
+                            x4, y4]
+
+            x1, y1, x2, y2, x3, y3, x4, y4 = polygons[c]
+            polygons[c] = [x1, y1,
+                            x2, y2,
+                            x3, y3,
+                            x4 + dx, y4 + dy]
+
+            x1, y1, x2, y2, x3, y3, x4, y4 = polygons[d]
+            polygons[d] = [x1 + dx, y1 + dy,
+                            x2, y2,
+                            x3, y3,
+                            x4, y4]
+
+        generated_mesh = []
+        for i in range(len(dimensions)):
+            generated_mesh.append([dimensions[i], polygons[i]])
+
         def do(image):
 
-            for a, b, c, d in polygon_indices:
-                dx = random.randint(-self.magnitude, self.magnitude)
-                dy = random.randint(-self.magnitude, self.magnitude)
-
-                x1, y1, x2, y2, x3, y3, x4, y4 = polygons[a]
-                polygons[a] = [x1, y1,
-                               x2, y2,
-                               x3 + dx, y3 + dy,
-                               x4, y4]
-
-                x1, y1, x2, y2, x3, y3, x4, y4 = polygons[b]
-                polygons[b] = [x1, y1,
-                               x2 + dx, y2 + dy,
-                               x3, y3,
-                               x4, y4]
-
-                x1, y1, x2, y2, x3, y3, x4, y4 = polygons[c]
-                polygons[c] = [x1, y1,
-                               x2, y2,
-                               x3, y3,
-                               x4 + dx, y4 + dy]
-
-                x1, y1, x2, y2, x3, y3, x4, y4 = polygons[d]
-                polygons[d] = [x1 + dx, y1 + dy,
-                               x2, y2,
-                               x3, y3,
-                               x4, y4]
-
-            generated_mesh = []
-            for i in range(len(dimensions)):
-                generated_mesh.append([dimensions[i], polygons[i]])
-
             return image.transform(image.size, Image.MESH, generated_mesh, resample=Image.BICUBIC)
 
         augmented_images = []
@@ -1873,3 +2028,102 @@ class ZoomGroundTruth(Operation):
             augmented_images.append(do(image))
 
         return augmented_images
+
+
+class Mixup(Operation):
+    """
+    Implements the *mixup* augmentation method, as described in:
+
+    Zhang et al. (2018), *mixup*: Beyond Empirical Risk Minimization,
+    arXiv:1710.09412
+
+    See `http://arxiv.org/abs/1710.09412 <http://arxiv.org/abs/1710.09412>`_
+    for details.
+
+    Also see `https://github.com/facebookresearch/mixup-cifar10 <https://github.com/facebookresearch/mixup-cifar10>`_
+    for code which was followed to create this functionality in Augmentor.
+
+    The *mixup* augmentation technique trains a neural network on
+    "*combinations of pairs of examples and their labels*" (Zhang et al., 2018).
+
+    In summary, *mixup* constructs training samples as follows:
+
+    .. math::
+        \\tilde{x} = \\lambda x_i + (1-\\lambda)x_j
+    .. math::
+        \\tilde{y} = \\lambda y_i + (1-\\lambda)y_j
+
+    where :math:`(x_i, y_i)` and :math:`(x_j, y_j)` are two samples from
+    the training data, :math:`x_i, x_j` are raw input vectors,
+    :math:`y_i, y_j` are one-hot label encodings (such as [0.0, 1.0]), and
+    :math:`\\lambda \\in [0, 1]` where :math:`\\lambda` is sampled randomly
+    from the Beta distribution, :math:`\\beta(\\alpha, \\alpha)`.
+
+    The :math:`\\alpha` hyper-parameter controls the strength of the
+    interpolation between image-label pairs, where
+    :math:`\\alpha \in \\{0, \\infty \\}`
+
+    According to the paper referenced above, values for :math:`\\alpha`
+    between 0.1 and 0.4 led to best performance. Smaller values for
+    :math:`\\alpha` result in less *mixup* effect whereas larger values would
+    tend to result in overfitting.
+    """
+    def __init__(self, probability, alpha=0.4):
+        """
+
+        .. note:: Not yet enabled!
+            This function is currently implemented but not **enabled**, as it
+            requires each image's label in order to operate - something which
+            Augmentor was not designed to handle.
+
+        :param probability: Controls the probability that the operation is
+         performed when it is invoked in the pipeline.
+        :param alpha: The alpha parameter controls the strength of the
+         interpolation between image-label pairs. It's value can be any value
+         greater than 0. A smaller value for :attr:`alpha` results in more
+         values closer to 0 or 1, meaning the *mixup* is more often closer to
+         either of the images in the pair. Its value is set to 0.4 by default.
+        :type probability: Float
+        :type alpha: Float
+        """
+        Operation.__init__(self, probability)
+        self.alpha = alpha
+
+    def perform_operation(self, images):
+        """
+        This function is currently implemented but not **enabled**, as it
+        requires each image's label in order to operate - something which
+        Augmentor was not designed to handle.
+
+        This is therefore future work, and may only be possible when used in
+        combination with generators.
+        """
+
+        if self.alpha > 0:
+            lambda_value = np.random.beta(self.alpha, self.alpha)
+        else:
+            lambda_value = 1
+
+        def do(image1, image2, y1, y2):
+
+            image1 = np.asarray(image1)
+            image1 = image1.astype('float32')
+
+            image2 = np.asarray(image2)
+            image2 = image2.astype('float32')
+
+            mixup_x = lambda_value * image1 + (1 - lambda_value) * image2
+
+            mixup_y = lambda_value * y1 + (1 - lambda_value) * y2
+
+            return mixup_x, mixup_y
+
+        augmented_images = []
+
+        y1 = np.array([0.0, 1.0])
+        y2 = np.array([1.0, 0.0])
+
+        for image in images:
+            augmented_images.append(do(image, image, y1, y2))
+
+        return augmented_images
