@@ -11,10 +11,10 @@ error, more reproducible, more efficient, and easier to perform.
 
 """
 
-from .Pipeline import Pipeline, DataFramePipeline
+from .Pipeline import Pipeline, DataFramePipeline, DataPipeline
 
 __author__ = """Marcus D. Bloice"""
 __email__ = 'marcus.bloice@medunigraz.at'
-__version__ = '0.2.1'
+__version__ = '0.2.9'
 
-__all__ = ['Pipeline', 'DataFramePipeline']
+__all__ = ['Pipeline', 'DataFramePipeline', 'DataPipeline']
